/**
 *	comsubs  --  communication subroutines for server side
 *		Copyright (C) Kazuro Furukawa, 1988-2004
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa  Nov.9.1988. (rwrited)
 *	K.F  Apr.13.1989.  converted from rwrited (rechod)
 *	K.F  Jun.17.1992.  rechod v3.4.2
 *	K.F  Jul.20.1992.(x)  comsubs
 *	Modifications
 *	K.F  Sep.14.1992.    add option 3 (bit 2)
 *	K.F  Oct.25.1992.(2) add reuseaddr in s2_sopen
 *	K.F  Nov.26.1992.    close failed socket correctly
 *	K.F  Mar.2.1993.     move error code definition to comsub.h
 *	K.F  Jun.15.1993.    add s2_upeername
 *	K.F  Aug.9.1994.     add s2_upeerport
 *	K.F  Feb.21.1996.    ansi C
 *	K.F  Dec.22.1997.    use sigaction() & waitpid(), SIGACTION
 *	K.F  Feb.15.1998.    add s2_peerinfo
 *	K.F  May.11.1999.    for NONETDB
 *	K.F  May.2.2011.     add reuseport in s2_sopen
 *
 *	This module is a re-writing of zechosub.
 *	This module defines open(server), accept and close rouines 
 *	for socket communication at server side.
 *	 int s2_sopen( portnum, service, protocol, option )
 *	 int s2_accept( sl, flag, option )
 *	 int s2_sclose( sd, option )
 *	Routines s2_receive and s2_send in comsub.c will be used for 
 *	"tcp" data transfer.
 *	For "udp" server,  use routines s2_ureceive and s2_usend 
 *	in this file, instead of s2_receive and s2_send.
 *	 int s2_ureceive( sl, buffer, buflen, option )
 *	 int s2_usend( sl, buffer, buflen, option )
 *	 char * s2_upeername( sl, option )
 *	 int s2_upeerport( sl, option )
 *	
 *	If option is 1, error message is printed on to stderr 
 *	when error occured.
 *	If option is 3, error and information message is printed on to 
 *	stderr when error occured.
 *	Error codes errno returned from the system subroutines are 
 *	preserved on return to the caller.
 *
 *	Bugs:
 *	As global data will be used between s2_ureceive and s2_usend
 *	in order to keep client's address, these routines must be paired
 *	to distinguish the clients.
 *	If the file descriptor 0 is closed between s2_sopen and s2_accept,
 *	s2_accept may return 0 for the connection file descriptor.
 */

static char	comsubs_what[] = 
"@(#)comsubs.c v1.8.0 tcp|udp server subroutines K.Furukawa Nov.1988-May.2011"; /*
**/

#include "comsub.h"			/* comsub definitions */
#include "comsub_p.h"			/* comsub private definitions */

static int	signalflag;		/* flag for SIGCLD signal */
/*static*/
 struct  sockaddr_in  s2_from;		/* socket address structure */
static int	s2_fromlen;

/**
 *  s2_sopen
 *	s2_sopen prepares communication server socket to accept a 
 *	a connection from a client.
 *
 *    int s2_sopen( portnum, service, protocol, option )
 *
 *	if portnum is positive, try to open a server socket with protocol
 *	then try to open with service name
 *	then if portnum is negative, try to open with -portnum
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than tcp/udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_BINDERR	socket bind failure
 *		-7 S2_CONNERR	socket connect failure
 *		-8 S2_LISNERR	socket listen failure
 */
int  s2_sopen( 
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  char *protocol,		/* "tcp" or "udp" */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    int	sl;					/* socket desctiptor */
    struct  sockaddr_in  scin;			/* socket address structure */
    int	usetcp;					/* flag for protocol tcp/udp */
    char *nm = "s2_sopen";			/* name of the routine */

    memset( (char *)&scin, 0, sizeof(scin) );	/* initialize data structure */

/* port/service handling */

    if ( !strcmp( protocol, "tcp" ) )  {
	usetcp = 1;				/* tcp protocol specified */
    }  else  {
	if ( !strcmp( protocol, "udp" ) )  {
	    usetcp = 0;				/* udp protocol specified */
	}  else  {
	    if ( option & S2_OALL )
		s2_errmsg( 0, nm, "protocol other than tcp/udp specified" );
	    return S2_NOPROT;			/* bad protocol name */
	}
    }

    if ( portnum > 0 )  {			/* port number specified */
	scin.sin_port = htons( portnum );	/* set service port number */
    }  else if ( service && strlen( service ) )  {
#ifndef NONETDB
	struct  servent  *sp;			/* service data structure */
	sp = getservbyname( service, protocol ); /* get service information */
	if ( sp )  {
	    scin.sin_port = sp->s_port;  	/* set port number */
	}  else  
#endif /* NONETDB */
	{
	    int port;
	    if ( *service == '#' )  service++;
	    if ( (port = atoi(service)) > 0 )  {
		scin.sin_port = htons( port );  /* port number in string */
	    }  else if ( portnum < 0 )  {
		scin.sin_port = htons( -portnum );  /* default port */
	    }  else  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "service not found\n" );
		return S2_NOSERV;		/* service not found */
	    }
	}
    }  else if ( portnum < 0 )  {
	scin.sin_port = htons( -portnum );	/* default port */
    }  else  {
	if ( option & S2_OALL )
	    s2_errmsg( 0, nm, "no port/service specified\n" );
	return S2_NOPORT;			/* no port/service specified */
    }

/* socket handling */

    scin.sin_family = AF_INET;
    scin.sin_addr.s_addr = INADDR_ANY;		/* listen to any address */

    sl = socket( AF_INET, ( usetcp ? SOCK_STREAM : SOCK_DGRAM ), 0 );
						/* create the listen socket */
    if ( sl == -1 )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "socket open failure" );
	return S2_OPENERR;			/* socket open failure */
    }

/* bind */

    {
	int iloop = 0;				/* bind listen to the socket */

#ifdef SO_REUSEADDR
	/* this makes port re-usable without delay, this does not work well */
	/* i don't know why now */
	{
	    /* long	reuseaddr = 1; */	/* reuse address on */
	    int	reuseaddr = 1;			/* reuse address on */
	    if ( setsockopt( sl, SOL_SOCKET, SO_REUSEADDR, (char *)&reuseaddr,
		 sizeof( reuseaddr ) ) == -1 )  {  /* set reuse addr */
		if ( option & S2_OINFO )
		    s2_errmsg( 1, nm, "warning - setsockopt(r) failure" );
	    }
	}
#endif

#ifdef SO_REUSEPORT
	/* this makes port re-usable for udp */
	{
	    int	reuseport = 1;			/* reuse port on */
	    if ( setsockopt( sl, SOL_SOCKET, SO_REUSEPORT, (char *)&reuseport,
		 sizeof( reuseport ) ) == -1 )  {  /* set reuse port */
		if ( option & S2_OINFO )
		    s2_errmsg( 1, nm, "warning - setsockopt(p) failure" );
	    }
	}
#endif

	while ( bind( sl, (void *)&scin, sizeof( scin ) ) == -1 )  {
	    if ( ++iloop > 4 )  {		/* if time out */
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "socket bind failure" );
		close( sl );
		return S2_UBNDERR;		/* socket bind failure */
	    }
	    s2_sleep( 5 );
	}
    }

    if ( usetcp )  {
	if ( listen( sl, 5 ) == -1 )  {	/* initiate the listen */
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "socket listen failure" );
	    close( sl );
	    return S2_LISNERR;			/* socket listen failure */
	}
    }

    return sl;					/* successful listen */
}


/**
 *  s2_accept
 *	s2_accept accepts a request from a client on the socket created
 *	by s2_sopen call and establishes a connection.  If multiple
 *	thread is specified, it will fork child.
 *
 *    int s2_accept( sl, flag, option )
 *
 *	sl is socket descriptor returned from s2_sopen.
 *	if flag is 1 multiple thread will be generated and child process
 *	accepts a connection.  otherwise only one connection at a time
 *	will be generated.
 *	(on os9/msdos/vms flag has no effect, always single thread.)
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	0	success, parent process return
 *    		positive	success, socket descriptor (for child process)
 *		-9  S2_ACCPERR	socket accept failure
 *		-10 S2_FORKERR	child fork failure
 */
int  s2_accept( 
  int sl,			/* socket descriptor for listen */
  int flag,			/* flag 1: multi-thread, 0: single thread */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    int	sn;					/* socket desctiptor */
    struct	sockaddr_in	from;		/* socket address structure */
    int	fromlen;
    char *nm = "s2_accept";			/* name of the routine */

#ifndef SINGLESERVER
    if ( flag == 1 )  {				/* ignore child-quit signals */
#ifdef SIGACTION
	void s2_ignorechild( int );		/* defined in this file */
	struct sigaction sa;			/* signal action spec. */
	memset( (void *)&sa, 0, sizeof(sa) );
	sa.sa_handler = s2_ignorechild;
	sa.sa_mask = sigmask(SIGCHLD);
	    
	sigaction( SIGCHLD, &sa, (struct sigaction *) NULL );
#else /* SIGACTION */
	void s2_ignorechild( void );		/* defined in this file */
	signal( SIGCLD, SIGNAL_CAST s2_ignorechild );/* not to leave zombies */
#endif /* SIGACTION */
    }
#endif

    signalflag = 0;				/* reset SIGCLD flag */
    fromlen = sizeof( from );
    
    while ( -1 == ( sn = accept( sl, (void *)&from, &fromlen ) ) )  {
						/* accept connection */
				/* on unix/vms "from" has enough info of
				   client, but on msdos/os9. */
	if ( !signalflag )  {			/* if SIGCLD, try again */
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "socket accept failure" );
	    return S2_ACCPERR;			/* socket accept failure */
	}
	signalflag = 0;				/* reset SIGCLD flag */
    }

#ifndef SINGLESERVER
    if ( flag == 1 )  {
	switch ( fork() )  {
	case -1:
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "child fork failure" );
	    return S2_FORKERR;			/* fork failure */

	case 0:					/* child process */
	    close( sl );			/* close listen socket */
	    break;

	default:				/* parent daemon */
	    close( sn );			/* close new socket */
	    return 0;
	}
    }
#endif

    if ( option & S2_OINFO )  {
	char	*hostname;
	int	port;
	char	*s2_currtime();			/* current time */

	if ( !s2_peerinfo( sn, &hostname, &port, option & ~S2_OINFO ) )  {
	    fprintf( stderr, " Connected from %s on remote port %u at %s\n", 
		    hostname, port, s2_currtime() );
	}
    }

    {
#ifndef NOLINGERTIME
	struct	linger solinger;		/* linger on, 10 seconds */
	solinger.l_onoff = 1;
	solinger.l_linger = 10;
#else
	long	solinger = 1;			/* linger on */
#endif
	if ( setsockopt( sn, SOL_SOCKET, SO_LINGER, (char *)&solinger, 
		sizeof( solinger ) ) == -1 )  { /* set graceful close */
	    if ( option & S2_OINFO )
		s2_errmsg( 1, nm, "warning - setsockopt(l) failure" );
	}
    }
    {
	long	sokeepalive = 1;		/* keepalive on */
	if ( setsockopt( sn, SOL_SOCKET, SO_KEEPALIVE, (char *)&sokeepalive,
			sizeof( sokeepalive ) ) == -1 )  {  /* set keepalive*/
	    if ( option & S2_OINFO )
		s2_errmsg( 1, nm, "warning - setsockopt(k) failure" );
	}
    }

    return sn;					/* return accepted socket */
}


/**
 *  s2_sclose
 *	s2_sclose close communication port of server side
 *
 *    int  s2_sclose( sd, option )
 *
 *    return:	nothing
 */
int  s2_sclose( 
  int sd,			/* socket descriptor returned from s2_open */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_sopen";			/* name of the routine */

/* information */

    if ( option & S2_OINFO )  {
	int	port;
	char	*s2_currtime();			/* current time */

	if ( !s2_peerinfo( sd, NULL, &port, option & ~S2_OINFO ) )  {
	    fprintf( stderr, " Disconnected remote port %u at %s\n", 
		    port, s2_currtime() );
	}
    }

#ifndef vms
    close( sd );				/* close socket */
#else
    netclose( sd );
#endif
    return 0;
}


/**
 *  s2_ureceive
 *	s2_ureceive receives message from "udp" communication port
 *	calls recvfrom(2) system call.
 *
 *    int  s2_ureceive(sl, buffer, buflen, option)
 *
 *    return:	-22 S2_RECVERR	receive failure
 *		positive	byte count which has been received
 */
int  s2_ureceive(
  int sl,			/* socket descriptor returned from s2_sopen */
  char *buffer,			/* buffer to receive */
  int buflen,			/* buffer length */
  int option			/* option 1: error/information message */
)
/*
**/
{
    char *nm = "s2_ureceive";			/* name of the routine */
    int iret;

    s2_fromlen = sizeof( s2_from );

    if (-1 == (iret = recvfrom( sl, buffer, buflen, 0, 
			       (void *)&s2_from, &s2_fromlen) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "recvfrom failure" );
	return S2_RECVERR;			/* receive failure */
    }
    return iret;
}


/**
 *  s2_usend
 *	s2_usend sends message to "udp" communication port
 *	calls sendto(2) system call.
 *
 *    int  s2_usend(sl, buffer, buflen, option)
 *
 *    return:	-21 S2_SENDERR	send failure
 *		positive	byte count which has been sent
 */
int  s2_usend( 
  int sl,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to send */
  int buflen,			/* buffer length */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_usend";			/* name of the routine */
    int iret;

    if (buflen > (iret = sendto(sl, buffer, buflen, 0, 
			     (void *)&s2_from, s2_fromlen) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "sendto failure" );
	return S2_SENDERR;			/* send failure */
    }
    return iret;
}


/**
 *  s2_peerinfo
 *	s2_peerinfo returns the remote host name and the port number 
 *	from which the "tcp" message is received.  this function should 
 *	be called after s2_accept() call and before s2_close().
 *
 *    char *  s2_peerinfo(sn, hostname, port, option)
 *
 *    return:	0		success
 *		-2  S2_NOPORT	failure
 *	hostname:	pointer to host name in temporarily allocated buffer
 *			if hostname is NULL, hostname is not returned
 *	port:		pointer to port number in local format
 *
 *	if option is S2_OINFO, it prints remote information.
 */
int s2_peerinfo( 
  int sn,			/* socket descriptor, not used */
  char **hostname,		/* pointer to host name to be returned */
  int *port,			/* port number to be returned */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_peerinfo";			/* name of the routine */

    struct  sockaddr_in  name;			/* socket for communication */
    int	namelen;

    namelen = sizeof( name );
    if ( getpeername(sn, (void *)&name, &namelen) == -1 )  {
						/* get client info */
	if ( option & S2_OERR )
	    s2_errmsg( 1, nm, "warning - getpeername failure" );
	return S2_NOPORT;			/* failure */

    }  else  {
	if ( hostname )  {			/* if NULL not returned */
#ifndef NONETDB
	    struct	hostent	*hp;
	    hp = gethostbyaddr( (char *) &name.sin_addr, 
			       sizeof( struct in_addr ), name.sin_family );
						/* get client information */
	    if ( hp != NULL )  {
		*hostname = hp->h_name;		/* host name of client */
	    }  else  
#endif /* NONETDB */
	    {
		*hostname = inet_ntoa( name.sin_addr ); /* dot format */
	    }
	}
	*port = ntohs( name.sin_port );		/* remote port number */

	if ( option & S2_OINFO )  {
	    fprintf( stderr, "Remote host %s, remote port %u\n", 
		*hostname, *port );
	}
    }

    return 0;
}


/**
 *  s2_upeername
 *	s2_upeername returns the remote host name from which the "udp" 
 *	message is received.  this function should be called just after
 *	s2_ureceive() or s2_tureceive() call.
 *
 *    char *  s2_upeername(sn, option)
 *
 *    return:	host name or host address in dotted format
 *
 *	if option is S2_OINFO, it prints remote information.
 */
char *  s2_upeername( 
  int sn,			/* socket descriptor, not used */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_upeername";			/* name of the routine */

    char	*hostname = " ";
#ifndef NONETDB
    struct	hostent	*hp;

    hp = gethostbyaddr( (char *) &s2_from.sin_addr, 
		       sizeof( struct in_addr ), s2_from.sin_family );
						/* get client information */
    if ( hp != NULL )  {
	hostname = hp->h_name;			/* host name of client */
    }  else  
#endif /* NONETDB */
    {
	hostname = inet_ntoa( s2_from.sin_addr ); /* dot format */
    }
    
    if ( option & S2_OINFO )
	fprintf( stderr, " Received from %s on remote port %u at %s\n",
		hostname, ntohs( s2_from.sin_port ), s2_currtime() );
    return hostname;
}


/**
 *  s2_upeerport
 *	s2_upeerport returns the remote port number from which the "udp" 
 *	message is received.  this function should be called just after
 *	s2_ureceive() or s2_tureceive() call.
 *
 *    int  s2_upeerport(sn, option)
 *
 *    return:	remote port number in operation
 *
 *	if option is S2_OINFO, it prints remote information.
 */
int  s2_upeerport( 
  int sn,			/* socket descriptor, not used */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_upeerport";			/* name of the routine */
    int upeerport;

    upeerport = ntohs( s2_from.sin_port );
    if ( option & S2_OINFO )
	fprintf( stderr, " Received on remote port %u\n", upeerport );
    return upeerport;
}


#ifndef SINGLESERVER
/*
 *  s2_ignorechild
 *	ignorechild accepts and only ignores the signal that alarms 
 *	a child process exits.
 */
#ifdef SIGACTION
void  s2_ignorechild( int sig )
{
    while ( 0 < waitpid( (pid_t) -1, (int *) NULL, WNOHANG ) )  {
	/* (pid_t) -1 means any child */
	;					/* wait child to exit */
    }
#ifdef REINSTALLSIGNAL				/* signal reinstallation */
    {
	struct sigaction sa;			/* signal action spec. */
	memset( (void *)&sa, 0, sizeof(sa) );
	sa.sa_handler = s2_ignorechild;
	sa.sa_mask = sigmask(SIGCHLD);
	sigaction( SIGCHLD, sa, (struct sigaction *) NULL );
    }
#endif /* REINSTALLSIGNAL */
    signalflag++;				/* set signal flag */
}
#else /* SIGACTION */
void  s2_ignorechild( void )
{
    wait( (char*) 0 );				/* wait child to exit */
#ifdef REINSTALLSIGNAL				/* signal reinstallation */
    signal( SIGCLD, s2_ignorechild );		/* not to leave zombies */
#endif /* REINSTALLSIGNAL */
    signalflag++;				/* set signal flag */
}
#endif /* SIGACTION */
#endif /* SINGLESERVER */
