/**
 *	s2_rechod  --  remote echo server program
 *		Copyright (C) Kazuro Furukawa, 1988-1993
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *
 *	K.Furukawa  Nov.9.1988. (rwrited)
 *	K.F  Apr.13.1989.  converted from rwrited (rechod)
 *	K.F  Jun.17.1992.  rechod v3.4.2
 *	K.F  Oct.2.1992.   s2_rechod using s2_xxx routines
 *	Modifications
 *	K.F  Nov.6.1992.   single server exit fix
 *	K.F  Nov.26.1992.  use getopt routine
 *	K.F  Feb.21.1996.  ansi C
 *
 *      This program receives messages from s2_recho program and echos back
 *      the the same message to recho program.  This will be the typical
 *	example for the TCP|UDP/IP server-client model.
**/

char	what[] = 
"@(#)s2_rechod.c v1.3.1 tcp|udp/ip echo server, K.Furukawa, Nov.1988-Aug.2003";

#include "comsub.h"				/* network header files */

#define  SERVICENAME	"s2_recho"		/* service name */
#define  PORTNUMB	11061			/* default service port */
#define  BUFFERSIZE	16384			/* communication buffer size */

char	*nm;
char	buf[BUFFERSIZE];			/* communication buffer */
int	usetcp;					/* flag for protocol tcp/udp */
int	portnum;				/* port number */

/**
 *  main
 *	Main prepares a listen socket for the recho service and forks a daemon
 *	process in order to disconnect the control terminal.  
**/
int  main( int argc, char ** argv )
{
    void argument( int, char ** );
    void errormsg( char * );
    void tcpserver( int );
    void udpserver( int );
    int	sl;					/* socket desctiptor */

    nm = argv[0];				/* program name */
    usetcp = 1;					/* use tcp by default */
    portnum = 0;				/* port number */
    
    argument(argc, argv);			/* command line argument */
    
    if ( !portnum )  {				/* port number not specified */
	portnum = -PORTNUMB;			/* default port */
    }

#ifndef SINGLESERVER
    setpgrp();				/* disconnecting control terminal */

    switch ( fork() )  {
    case -1:
	errormsg( "daemon fork failure" );
	exit( 1 );
    case 0:					/* daemon */
#endif
	fclose( stdin );			/* close control terminal */
	sl = s2_sopen( portnum, SERVICENAME, (usetcp?"tcp":"udp"), S2_OALL );
	if ( sl<0 )  exit( 1 );			/* create listen socket */

	if ( usetcp )  tcpserver( sl );		/* server for tcp */
	else  udpserver( sl );			/* server for udp */

#ifndef SINGLESERVER
    default:					/* parent process */
	exit( 0 );				/* free terminal */
    }
#endif
    return 0;
}

/**
 *  tcpserver
 *	Daemon waits for connections from clients.  If a request arrives,
 *	it forks a child to process each request.
 *	Server reports the connection information to standard error (
 *	maybe logging file).  Then it echos the message received from 
 *	the client.  It also reports disconnection information.
**/
void  tcpserver( int sl )
{
    int	sn;					/* socket desctiptor */

    for(;;)  {
#ifndef SINGLESERVER
	sn = s2_accept( sl, 1, S2_OALL );	/* accept connection */
#else
	sn = s2_accept( sl, 0, S2_OALL );	/* accept connection */
#endif
	if (sn<0) exit(1);
	if (sn>0)  {
	    int	cc;
	    long lines = 0;			/* communication counter */
	    long bytes = 0;			/* communication counter */
						/* server main body starts */
	    while( ( cc = s2_receive(sn, buf, BUFFERSIZE, S2_OALL) ) > 0 )  {
						/* read message */
		if( s2_send(sn, buf, cc, S2_OALL) < cc )  {	/* echo back */
		    exit( 1 );
		}
		lines++;
		bytes += cc;
	    }

	    fprintf( stderr, "  %ld lines %ld bytes\n", lines, bytes );
	    s2_sclose( sn, S2_OALL );
#ifndef SINGLESERVER
	    exit( 0 );				/* child exit */
#endif
	}
    }						/* parent repeat forever */
}

/**
 *  udpserver
 *	Udpserver waits for message from clients.  If a message arrives,
 *	it echoes the message back to the sender.
**/
void  udpserver( int sl )
{
    int	cc;

    while( ( cc = s2_ureceive( sl, buf, BUFFERSIZE, S2_OALL ) ) >= 0 )  {
						/* read message */
	if( s2_usend(sl, buf, cc, S2_OALL) < cc ) {	/* echo back*/
	    exit( 1 );
	}

    }						/* repeat forever */
    exit( 1 );
}

/*
 *  errormsg
 *	print error message
 */
void  errormsg( char * msg )
{
	perror( nm );
	fprintf( stderr, "%s: %s\n", nm, msg );
}

/**
 *  argument
 *	Argument handles command line argument, such as hostname or options.
**/
void  argument( int argc, char ** argv )
{
	int c;
	extern int optind, opterr;
	extern char *optarg;
	opterr = 0;			/* disable informational message */

	while ((c = getopt (argc, argv, "uUp:P:hH")) != EOF)
	    switch ( c )  {
	    case 'u':
	    case 'U':				/* use udp */
		usetcp = 0;			/* reset tcp flag */
		break;
	    case 'p':
	    case 'P':				/* port number */
		portnum = atoi( optarg );
		break;
	    case '?':
	    case 'h':
	    case 'H':				/* help */
	    default:
		goto usage;			/* print out usage */
	    }
	if ( argc == optind )  return;

usage:
	fprintf(stderr, "%s\n", what );
	fprintf(stderr, "Usage:  s2_rechod  <option>\n" );
	fprintf(stderr, " option:  -u   use UDP instead of TCP\n" );
	fprintf(stderr, "          -pX  use service port number X\n" );
	fprintf(stderr, "          -h   help message\n" );
	fprintf(stderr, " example: s2_rechod \n" );
	fprintf(stderr, "          s2_rechod -u -p2000\n" );
	exit( 1 );
}
